{
  "type": "record",
  "name": "Struct",
  "namespace": "Avrotize",
  "fields": [
    {
      "name": "data",
      "type": {
        "type": "map",
        "values": [
          "null",
          "boolean",
          "int",
          "long",
          "float",
          "double",
          "bytes",
          "string",
          {
            "type": "array",
            "items": {
              "type": "record",
              "name": "StructNestedValue",
              "fields": [
                {
                  "name": "value",
                  "type": [
                    "null",
                    "boolean",
                    "int",
                    "long",
                    "float",
                    "double",
                    "bytes",
                    "string",
                    {
                      "type": "array",
                      "items": "StructNestedValue"
                    },
                    {
                      "type": "map",
                      "values": "StructNestedValue"
                    }
                  ]
                }
              ]
            }
          },
          {
            "type": "map",
            "values": "StructNestedValue"
          }
        ]
      }
    }
  ]
}
