{
    "compilerOptions": {
        "target": "ES2022",
        "module": "ES2022",
        "strict": true,
        "esModuleInterop": true,
        "skipLibCheck": true,
        "forceConsistentCasingInFileNames": true,
        "outDir": "./dist",
        "rootDir": "./src",
        "experimentalDecorators": true,
        "emitDecoratorMetadata": true,
        "declaration": true,
        "declarationMap": true,
        "moduleResolution": "node",
        "resolveJsonModule": true
    },
    "include": [
        "src/**/*"
    ]
}
