{
  "name": "{{ package_name }}",
  "version": "1.0.0",
  "description": "Generated JavaScript classes from JSON Structure schema",
  "main": "src/index.js",
  "scripts": {
    "test": "node test/test_runner.js"
  },
  "keywords": [],
  "author": "",
  "license": "ISC"
}
