{%- set slnguid = uuid() -%}
{%- set prjguid = uuid() -%}
{%- set testguid = uuid() -%}
{%- set slnguid2 = uuid() -%}
Microsoft Visual Studio Solution File, Format Version 12.00
# Visual Studio Version 17
VisualStudioVersion = 17.5.002.0
MinimumVisualStudioVersion = 10.0.40219.1
Project("{ {{-slnguid-}} }") = "{{project_name|pascal}}", "src\{{project_name|pascal}}.csproj", "{ {{-prjguid-}} }"
EndProject
Project("{ {{-slnguid-}} }") = "{{project_name|pascal}}_test", "test\{{project_name|pascal}}.Test.csproj", "{ {{-testguid-}} }"
EndProject
Global
	GlobalSection(SolutionConfigurationPlatforms) = preSolution
		Debug|Any CPU = Debug|Any CPU
		Release|Any CPU = Release|Any CPU
	EndGlobalSection
	GlobalSection(ProjectConfigurationPlatforms) = postSolution
		{ {{-prjguid-}} }.Debug|Any CPU.ActiveCfg = Debug|Any CPU
		{ {{-prjguid-}} }.Debug|Any CPU.Build.0 = Debug|Any CPU
		{ {{-prjguid-}} }.Release|Any CPU.ActiveCfg = Release|Any CPU
		{ {{-prjguid-}} }.Release|Any CPU.Build.0 = Release|Any CPU
		{ {{-testguid-}} }.Debug|Any CPU.ActiveCfg = Debug|Any CPU
		{ {{-testguid-}} }.Debug|Any CPU.Build.0 = Debug|Any CPU
		{ {{-testguid-}} }.Release|Any CPU.ActiveCfg = Release|Any CPU
		{ {{-testguid-}} }.Release|Any CPU.Build.0 = Release|Any CPU
	EndGlobalSection
	GlobalSection(SolutionProperties) = preSolution
		HideSolutionNode = FALSE
	EndGlobalSection
	GlobalSection(ExtensibilityGlobals) = postSolution
		SolutionGuid = { {{-slnguid2-}} }
	EndGlobalSection
EndGlobal