<Project Sdk="Microsoft.NET.Sdk">
    <PropertyGroup>
        <TargetFramework>net9.0</TargetFramework>
        <Nullable>enable</Nullable>
        <GenerateDocumentationFile>true</GenerateDocumentationFile>
    </PropertyGroup>
    <ItemGroup>
        {%- if avro_annotation %}
        <PackageReference Include="Apache.Avro" Version="{{ CSHARP_AVRO_VERSION }}" />
        {%- endif %}
        {%- if protobuf_net_annotation %}
        <PackageReference Include="protobuf-net" Version="{{ PROTOBUF_NET_VERSION }}" />
        {%- endif %}
        {%- if newtonsoft_json_annotation %}
        <PackageReference Include="Newtonsoft.Json" Version="{{ NEWTONSOFT_JSON_VERSION }}" />
        {%- endif %}
        {%- if system_text_json_annotation %}
        <PackageReference Include="System.Text.Json" Version="{{ SYSTEM_TEXT_JSON_VERSION }}" />
        {%- endif %}
        {%- if msgpack_annotation %}
        <PackageReference Include="MessagePack" Version="{{ MSGPACK_VERSION }}" />
        {%- endif %}
        <PackageReference Include="System.Memory.Data" Version="{{ SYSTEM_MEMORY_DATA_VERSION }}" />
    </ItemGroup>
    <ItemGroup>
        <PackageReference Include="NUnit" Version="{{ NUNIT_VERSION }}" />
        <PackageReference Include="NUnit3TestAdapter" Version="{{ NUNIT_ADAPTER_VERSION }}" />
        <PackageReference Include="Microsoft.NET.Test.Sdk" Version="{{ MSTEST_SDK_VERSION }}" />
    </ItemGroup>
</Project>  