<Project Sdk="Microsoft.NET.Sdk">
    <PropertyGroup>
        <TargetFramework>net9.0</TargetFramework>
        <Nullable>enable</Nullable>
        <GenerateDocumentationFile>true</GenerateDocumentationFile>
        <OutputType>Exe</OutputType>
        <StartupObject>InstanceSerializerProgram</StartupObject>
    </PropertyGroup>
    <ItemGroup>
        <ProjectReference Include="../src/{{project_name | pascal}}.csproj"/>
    </ItemGroup>
    <ItemGroup>
        <PackageReference Include="NUnit" Version="{{ NUNIT_VERSION }}" />
        <PackageReference Include="NUnit3TestAdapter" Version="{{ NUNIT_ADAPTER_VERSION }}" />
        <PackageReference Include="Microsoft.NET.Test.Sdk" Version="{{ MSTEST_SDK_VERSION }}" />
    </ItemGroup>
</Project>  

