{
  "name": "{{project_name}}",
  "version": "1.0.0",
  "dependencies": [
    "boost-uuid",
    "gtest"{% if json_annotation %},
    "nlohmann-json",
    "zlib",
    "boost-iostreams"{% endif %}
  ]
}
