[
  {
    "type": "record",
    "name": "Timestamp",
    "namespace": "google.protobuf",
    "fields": [
      {
        "name": "seconds",
        "type": "long",
        "doc": "Represents seconds of UTC time since Unix epoch 1970-01-01T00:00:00Z. Must be from 0001-01-01T00:00:00Z to 9999-12-31T23:59:59Z inclusive."
      },
      {
        "name": "nanos",
        "type": "int",
        "doc": "second values with fractions must still have non-negative nanos values that count forward in time. Must be from 0 to 999,999,999 inclusive."
      }
    ],
    "doc": "the Joda Time's [`ISODateTimeFormat.dateTime()`]( http:joda-time.sourceforge.net/apidocs/org/joda/time/format/ISODateTimeFormat.html#dateTime() ) to obtain a formatter capable of generating timestamps in this format."
  }
]